from pydantic import BaseModel, root_validator
from typing import Optional, Dict
from .quality_indicator_type import QualityIndicatorType
from .coverage_wfs import CoverageWfs


class QualityIndicator(BaseModel):
    type: QualityIndicatorType
    quality_dimension_id: str
    quality_dimension_name: str
    quality_warning_text: str
    warning_threshold: str
    property: Optional[str] = None
    input_filter: Optional[str] = None
    wfs: Optional[CoverageWfs] = None

    @root_validator(pre=False)
    def check_coverage(cls, values: Dict) -> Dict:
        type, wfs = values.get('type'), values.get('wfs')

        if type == QualityIndicatorType.COVERAGE and wfs is None:
            raise ValueError(
                'If the quality indicator type is "coverage", the property "wfs" must be set')

        return values



import re
from os import environ
from typing import List
from datetime import datetime, timezone
from lxml import etree as ET
from ...models.exceptions import DokAnalysisException


def get_env_var(var_name) -> str:
    try:
        return environ[var_name]
    except KeyError:
        raise DokAnalysisException('The environment variable ' + var_name + ' is not set')


def from_camel_case(value):
    regex = r"([A-Z])"
    subst = " \\1"
    result = re.sub(regex, subst, value, 0)

    return result.capitalize()


def to_camel_case(text: str) -> str:
    if text[0].islower():
        return text

    matches = re.findall('[A-ZÆØÅ][^A-ZÆØÅ]*', text)
    words: List[str] = list(map(lambda word: word.strip(' -_'), matches))
    result = words[0].lower() + ''.join(word.capitalize()
                                        for word in words[1:])

    return result


def keys_to_camel_case(data: dict) -> dict:
    return {to_camel_case(key): keys_to_camel_case(value) if isinstance(value, dict) else value for key, value in data.items()}


def parse_string(value: str) -> str | int | float | bool:
    if value is None:
        return None

    if value.isdigit():
        return int(value)
    elif value.replace('.', '', 1).isdigit() and value.count('.') < 2:
        return float(value)
    elif value.lower() == True:
        return True
    elif value.lower() == False:
        return False

    return value


def parse_date_string(value: str) -> datetime:
    try:
        return datetime.fromisoformat(value)
    except:
        return None


def should_refresh_cache(file_path: str, cache_days: int) -> bool:
    timestamp = file_path.stat().st_mtime
    modified = datetime.fromtimestamp(timestamp, tz=timezone.utc)
    diff = datetime.now(tz=timezone.utc) - modified

    return diff.days > cache_days


def xpath_select(element: ET._Element, path: str) -> any:
    return element.xpath(path)


def xpath_select_one(element: ET._Element, path: str) -> any:
    result = element.xpath(path)

    if len(result) == 0:
        return None

    if len(result) == 1:
        return result[0]

    raise Exception('Found more than one element')


def evaluate_condition(condition: str, data: dict[str, any]) -> bool:
    parsed_condition = _parse_condition(condition)
    result = eval(parsed_condition, data.copy())

    if isinstance(result, (bool)):
        return result

    raise Exception


def _parse_condition(condition: str) -> str:
    regex = r'(?<!=|>|<)\s*=\s*(?!=)'
    condition = re.sub(regex, ' == ', condition, 0, re.MULTILINE)

    return _replace_all(
        condition, {' AND ': ' and ', ' OR ': ' or ', ' IN ': ' in ', ' NOT ': ' not '})


def _replace_all(text: str, replacements: dict) -> str:
    for i, j in replacements.items():
        text = text.replace(i, j)
    return text


__all__ = [
    'background_tasks',
    'get_env_var', 
    'from_camel_case',
    'to_camel_case',
    'keys_to_camel_case',
    'parse_string',
    'parse_date_string',
    'should_refresh_cache',
    'xpath_select',
    'xpath_select_one',
    'evaluate_condition'
]


import logging
import yaml
from pathlib import Path
from typing import Dict, List, Tuple
from uuid import UUID
from pydantic import ValidationError
from cachetools import cached, TTLCache
from ..models.exceptions import DokAnalysisException
from ..models.config import DatasetConfig, QualityConfig, QualityIndicator
from ..utils.helpers.common import get_env_var

_LOGGER = logging.getLogger(__name__)


def get_dataset_configs() -> List[DatasetConfig]:
    dataset_configs, _ = _load_config()

    return dataset_configs


def get_dataset_config(dataset_id: UUID) -> DatasetConfig:
    dataset_configs, _ = _load_config()

    config = next(
        (conf for conf in dataset_configs if conf.dataset_id == dataset_id), None)

    return config


def get_quality_indicator_configs(dataset_id: UUID) -> List[QualityIndicator]:
    _, quality_configs = _load_config()
    indicators: List[QualityIndicator] = []

    for config in quality_configs:
        id: UUID = config.dataset_id

        if not id or id == dataset_id:
            indicators.extend(config.indicators)

    return indicators


@cached(cache=TTLCache(maxsize=1024, ttl=120))
def _load_config() -> Tuple[List[DatasetConfig], List[QualityConfig]]:
    config_dir = get_env_var('DOKANALYSE_CONFIG_DIR')

    if config_dir is None:
        raise DokAnalysisException(
            'The environment variable "DOKANALYSE_CONFIG_DIR" is not set')

    path = Path(config_dir)

    if not path.is_dir():
        raise DokAnalysisException(
            f'The "DOKANALYSE_CONFIG_DIR" path ({path}) is not a directory')

    glob = path.glob('*.yml')
    files = [path for path in glob if path.is_file()]

    if len(files) == 0:
        raise DokAnalysisException(
            f'The "DOKANALYSE_CONFIG_DIR" path ({path}) contains no YAML files')

    return _create_configs(files)


def _create_configs(files: List[Path]) -> Tuple[List[DatasetConfig], List[QualityConfig]]:
    dataset_configs: List[DatasetConfig] = []
    quality_configs: List[QualityConfig] = []

    for file_path in files:
        with open(file_path, 'r') as file:
            results = yaml.safe_load_all(file)
            result: dict

            for result in results:
                type = result.get('type')

                if type == 'dataset':
                    config = _create_dataset_config(result)

                    if config is not None:
                        dataset_configs.append(config)
                elif type == 'quality':
                    config = _create_quality_config(result)

                    if config is not None:
                        quality_configs.append(config)

    if len(dataset_configs) == 0:
        raise DokAnalysisException(
            f'Could not create any dataset configurations from the files in "DOKANALYSE_CONFIG_DIR"')

    return dataset_configs, quality_configs


def _create_dataset_config(data: Dict) -> DatasetConfig:
    try:
        config = DatasetConfig(**data)

        return config if not config.disabled else None
    except ValidationError as error:
        _LOGGER.error(error)
        return None


def _create_quality_config(data: Dict) -> QualityConfig:
    try:
        return QualityConfig(**data)
    except ValidationError as error:
        _LOGGER.error(error)
        return None


__all__ = [
    'get_dataset_configs',
    'get_dataset_config',
    'get_quality_indicator_configs'
]



from os import path
import json
from uuid import UUID
from pathlib import Path
from typing import List, Dict, Tuple
import aiohttp
from ..utils.helpers.common import should_refresh_cache
from ..utils.constants import CACHE_DIR

_API_URL = 'https://register.geonorge.no/api/dok-statusregisteret.json'

_CACHE_DAYS = 7

_CATEGEORY_MAPPINGS = {
    'BuildingMatter': ('egnethet_byggesak', 'Byggesak'),
    'MunicipalLandUseElementPlan': ('egnethet_kommuneplan', 'Kommuneplan'),
    'ZoningPlan': ('egnethet_reguleringsplan', 'Reguleringsplan')
}

_VALUE_MAPPINGS = {
    0: 'Ikke egnet',
    1: 'Dårlig egnet',
    2: 'Noe egnet',
    3: 'Egnet',
    4: 'Godt egnet',
    5: 'Svært godt egnet'
}


async def get_dok_status_for_dataset(dataset_id: UUID) -> Dict:
    dok_status_all = await get_dok_status()

    for dok_status in dok_status_all:
        if dok_status.get('dataset_id') == str(dataset_id):
            return dok_status

    return None


async def get_dok_status() -> List[Dict]:
    file_path = Path(
        path.join(CACHE_DIR, 'dok-status.json'))

    if not file_path.exists() or should_refresh_cache(file_path, _CACHE_DAYS):
        file_path.parent.mkdir(parents=True, exist_ok=True)
        dok_status = await _get_dok_status()
        json_object = json.dumps(dok_status, indent=2)

        with file_path.open('w', encoding='utf-8') as file:
            file.write(json_object)

        return dok_status
    else:
        with file_path.open(encoding='utf-8') as file:
            dok_status = json.load(file)

        return dok_status


async def _get_dok_status() -> List[Dict]:
    response = await _fetch_dok_status()

    if response is None:
        return []

    contained_items: List[Dict] = response.get('containeditems', [])
    datasets: List[Dict] = []

    for item in contained_items:
        dataset_id = _get_dataset_id(item)
        categories = _get_relevant_categories(item)
        suitability = []

        for key, value in categories:
            id, name = _CATEGEORY_MAPPINGS.get(key)

            suitability.append({
                'quality_dimension_id': id,
                'quality_dimension_name': name,
                'value': value,
                'comment': _VALUE_MAPPINGS.get(value)
            })

        datasets.append({
            'dataset_id': dataset_id,
            'suitability': suitability
        })

    return datasets


async def _fetch_dok_status() -> Dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(_API_URL) as response:
                if response.status != 200:
                    return None

                return await response.json()
    except:
        return None


def _get_dataset_id(item) -> List[Tuple]:
    metadata_url: str = item.get('MetadataUrl')
    dataset_id = metadata_url.split('/')[-1]

    return dataset_id


def _get_relevant_categories(item) -> List[Tuple]:
    suitability: Dict = item.get('Suitability')
    categories = [(key, value) for key, value in suitability.items()
                  if key in _CATEGEORY_MAPPINGS.keys()]

    return categories


__all__ = ['get_dok_status_for_dataset', 'get_dok_status']



from typing import Dict


class QualityMeasurement:
    def __init__(self, quality_dimension_id: str, quality_dimension_name: str, value: str | int | float | bool, comment: str):
        self.quality_dimension_id = quality_dimension_id
        self.quality_dimension_name = quality_dimension_name
        self.value = value
        self.comment = comment

    def to_dict(self) -> Dict:
        return {
            'qualityDimensionId': self.quality_dimension_id,
            'qualityDimensionName': self.quality_dimension_name,
            'value': self.value,
            'comment': self.comment
        }



from sys import maxsize
import json
from uuid import UUID
from typing import List, Dict
from osgeo import ogr
from .analysis import Analysis
from .result_status import ResultStatus
from .config.dataset_config import DatasetConfig
from ..services.geolett import get_geolett_data
from ..services.raster_result import get_wms_url, get_cartography_url
from ..utils.helpers.geometry import create_buffered_geometry, geometry_from_json
from ..http_clients.arcgis import query_arcgis


class ArcGisAnalysis(Analysis):
    def __init__(self, dataset_id: UUID, config: DatasetConfig, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int):
        super().__init__(dataset_id, config, geometry, epsg, orig_epsg, buffer)

    async def _run_queries(self) -> None:
        first_layer = self.config.layers[0]
        geolett_data = await get_geolett_data(first_layer.geolett_id)
        self._add_run_algorithm(f'query {self.config.arcgis}')             

        for layer in self.config.layers:
            if layer.filter is not None:
                self._add_run_algorithm(f'add filter {layer.filter}')

            status_code, api_response = await query_arcgis(
                self.config.arcgis, layer.arcgis, layer.filter, self.run_on_input_geometry, self.epsg)

            if status_code == 408:
                self.result_status = ResultStatus.TIMEOUT
                self._add_run_algorithm(f'intersects layer {layer.arcgis} (Timeout)')     
                break
            elif status_code != 200:
                self.result_status = ResultStatus.ERROR
                self._add_run_algorithm(f'intersects layer {layer.arcgis} (Error)')
                break

            if api_response:
                response = self.__parse_response(api_response)

                if len(response['properties']) > 0:
                    self._add_run_algorithm(f'intersects layer {layer.arcgis} (True)')
                    geolett_data = await get_geolett_data(layer.geolett_id)

                    self.data = response['properties']
                    self.geometries = response['geometries']
                    self.raster_result_map = get_wms_url(
                        self.config.wms, layer.wms)
                    self.cartography = await get_cartography_url(
                        self.config.wms, layer.wms)
                    self.result_status = layer.result_status
                    break

                self._add_run_algorithm(f'intersects layer {layer.arcgis} (False)')

        self.geolett = geolett_data

    async def _set_distance_to_object(self) -> None:
        buffered_geom = create_buffered_geometry(
            self.geometry, 20000, self.epsg)
        layer = self.config.layers[0]

        _, response = await query_arcgis(self.config.arcgis, layer.arcgis, layer.filter, buffered_geom, self.epsg)

        if response is None:
            self.distance_to_object = maxsize
            return

        distances = []

        for feature in response['features']:
            feature_geom = self.__get_geometry_from_response(feature)

            if feature_geom is not None:
                distance = round(
                    self.run_on_input_geometry.Distance(feature_geom))
                distances.append(distance)

        distances.sort()
        self._add_run_algorithm('get distance to nearest object')

        if len(distances) == 0:
            self.distance_to_object = maxsize
        else:
            self.distance_to_object = distances[0]

    def __parse_response(self, arcgis_response: Dict) -> Dict[str, List]:
        data = {
            'properties': [],
            'geometries': []
        }

        features: List[Dict] = arcgis_response.get('features', [])

        for feature in features:
            data['properties'].append(
                self.__map_properties(feature, self.config.properties))
            data['geometries'].append(
                self.__get_geometry_from_response(feature))

        return data

    def __map_properties(self, feature: Dict, mappings: List[str]) -> Dict:
        props = {}
        feature_props: Dict = feature['properties']

        for mapping in mappings:
            props[mapping] = feature_props.get(mapping)

        return props

    def __get_geometry_from_response(self, feature) -> ogr.Geometry:
        json_str = json.dumps(feature['geometry'])

        return geometry_from_json(json_str)



from io import BytesIO
from typing import List, Tuple
from lxml import etree as ET
from osgeo import ogr
from ..models.config import CoverageWfs
from ..http_clients.wfs import query_wfs
from ..utils.helpers.common import xpath_select_one
from ..utils.helpers.geometry import geometry_from_gml


async def get_values_from_wfs(wfs_config: CoverageWfs, geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float]:
    _, response = await query_wfs(wfs_config.url, wfs_config.layer, wfs_config.geom_field, geometry, epsg)

    if response is None:
        return [], 0

    source = BytesIO(response.encode('utf-8'))
    context = ET.iterparse(source, huge_tree=True)

    prop_path = f'.//*[local-name() = "{wfs_config.property}"]/text()'
    geom_path = f'.//*[local-name() = "{wfs_config.geom_field}"]/*'
    values: List[str] = []
    feature_geoms: List[ogr.Geometry] = []
    hit_area_percent = 0

    for _, elem in context:
        localname = ET.QName(elem).localname

        if localname == 'member':
            value = xpath_select_one(elem, prop_path)
            values.append(value)

            if value == 'ikkeKartlagt':
                geom_element = xpath_select_one(elem, geom_path)
                gml_str = ET.tostring(geom_element, encoding='unicode')
                feature_geom = geometry_from_gml(gml_str)

                if feature_geom:
                    feature_geoms.append(feature_geom)

    if len(feature_geoms) > 0:
        hit_area_percent = _get_hit_area_percent(geometry, feature_geoms)

    distinct_values = list(set(values))

    return distinct_values, hit_area_percent


def _get_hit_area_percent(geometry: ogr.Geometry, feature_geometries: List[ogr.Geometry]) -> float:
    geom_area: float = geometry.GetArea()
    hit_area: float = 0

    for geom in feature_geometries:
        intersection: ogr.Geometry = geom.Intersection(geometry)

        if intersection is None:
            continue

        area: float = intersection.GetArea()
        hit_area += area

    percent = (hit_area / geom_area) * 100

    return round(percent, 2)


__all__ = ['get_values_from_wfs']



from .coverage_wfs import CoverageWfs
from .dataset_config import DatasetConfig
from .layer import Layer
from .quality_config import QualityConfig
from .quality_indicator import QualityIndicator
from .quality_indicator_type import QualityIndicatorType


from pydantic import BaseModel, root_validator, validator
from typing import List, Optional, Dict
from ..result_status import ResultStatus
import uuid


class Layer(BaseModel):
    wfs: Optional[str] = None
    arcgis: Optional[str] = None
    ogc_api: Optional[str] = None
    wms: List[str]
    filter: Optional[str] = None
    result_status: ResultStatus
    geolett_id: Optional[uuid.UUID] = None

    @validator('result_status')
    def check_result_status(cls, value: ResultStatus) -> ResultStatus:
        valid_statuses = [
            ResultStatus.NO_HIT_GREEN,
            ResultStatus.NO_HIT_YELLOW,
            ResultStatus.HIT_YELLOW,
            ResultStatus.HIT_RED
        ]

        if value not in valid_statuses:
            raise ValueError('The layer\'s result_status must be either ' + 
                             ', '.join(list(map(lambda status: status.value, valid_statuses))))

        return value

    @root_validator(pre=True)
    def check_layer_type(cls, values: Dict) -> Dict:
        if not 'wfs' in values and not 'arcgis' in values and not 'ogc_api' in values:
            raise ValueError(
               'The layer must have either the "wfs", "arcgis" or "ogc_api" property set')

        return values



from enum import Enum

class ResultStatus(str, Enum):
    NO_HIT_GREEN = 'NO-HIT-GREEN'
    NO_HIT_YELLOW = 'NO-HIT-YELLOW'
    HIT_YELLOW = 'HIT-YELLOW'
    HIT_RED = 'HIT-RED'
    NOT_RELEVANT = 'NOT-RELEVANT'
    TIMEOUT = 'TIMEOUT'
    ERROR = 'ERROR'


from typing import List
from .legend import create_legend


def get_wms_url(wms_url: str, wms_layers: List[str]) -> str:
    layers = ','.join(wms_layers)

    return f'{wms_url}?layers={layers}'


async def get_cartography_url(wms_url: str, wms_layers: List[str]) -> str:
    urls = []

    for wms_layer in wms_layers:
        # autopep8: off
        url = f'{wms_url}?service=WMS&version=1.3.0&request=GetLegendGraphic&sld_version=1.1.0&layer={wms_layer.strip()}&format=image/png'
        # autopep8: on
        urls.append(url)

    if len(urls) == 1:
        return urls[0]

    data_url = await create_legend(urls)

    return data_url


__all__ = ['get_wms_url', 'get_cartography_url']



from contextvars import ContextVar
from flask_http_middleware import BaseHTTPMiddleware

_CORRELATION_ID_HEADER_NAME = 'x-correlation-id'

_correlation_id_ctx_var: ContextVar[str] = ContextVar(
    _CORRELATION_ID_HEADER_NAME, default=None)


def get_correlation_id() -> str:
    return _correlation_id_ctx_var.get()


class CorrelationIdMiddleware(BaseHTTPMiddleware):
    def __init__(self):
        super().__init__()

    def dispatch(self, request, call_next):
        header = request.headers.get(_CORRELATION_ID_HEADER_NAME)
        correlation_id = _correlation_id_ctx_var.set(header)

        response = call_next(request)
        _correlation_id_ctx_var.reset(correlation_id)

        return response


__all__ = ['CorrelationIdMiddleware', 'get_correlation_id']


from os import path
import json
from pathlib import Path
from uuid import UUID
from typing import Dict
import aiohttp
from ..models.metadata import Metadata
from ..utils.helpers.common import should_refresh_cache
from ..utils.constants import CACHE_DIR

_API_BASE_URL = 'https://kartkatalog.geonorge.no/api/getdata'
_CACHE_DAYS = 2

async def get_kartkatalog_metadata(metadata_id: UUID) -> Metadata:
    if metadata_id is None:
        return None

    metadata = await _get_kartkatalog_metadata(metadata_id)

    if metadata is None:
        return None

    return Metadata.from_dict(metadata)


async def _get_kartkatalog_metadata(metadata_id: UUID) -> Dict:
    file_path = Path(
        path.join(CACHE_DIR, f'kartkatalog/{metadata_id}.json'))

    if not file_path.exists() or should_refresh_cache(file_path, _CACHE_DAYS):
        file_path.parent.mkdir(parents=True, exist_ok=True)
        response = await _fetch_kartkatalog_metadata(metadata_id)

        if response is None:
            return None

        metadata = _map_response(metadata_id, response)
        json_object = json.dumps(metadata, indent=2)

        with file_path.open('w', encoding='utf-8') as file:
            file.write(json_object)

        return metadata
    else:
        with file_path.open(encoding='utf-8') as file:
            metadata = json.load(file)

        return metadata


def _map_response(metadata_id: UUID, response: Dict) -> Dict:
    title = response.get('NorwegianTitle')
    description = response.get('Abstract')
    owner = response.get('ContactOwner', {}).get('Organization')
    updated = response.get('DateUpdated')
    dataset_description_uri = 'https://kartkatalog.geonorge.no/metadata/' + str(metadata_id)

    return {
        'datasetId': str(metadata_id),
        'title': title,
        'description': description,
        'owner': owner,
        'updated': updated,
        'datasetDescriptionUri': dataset_description_uri
    }


async def _fetch_kartkatalog_metadata(metadata_id: UUID) -> Dict:
    try:
        url = f'{_API_BASE_URL}/{str(metadata_id)}'

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    return None

                return await response.json()
    except:
        return None


__all__ = ['get_kartkatalog_metadata']



from os import path
from pathlib import Path
import json
from uuid import UUID
from typing import List, Dict, Literal
import aiohttp
from ..models.config import DatasetConfig
from ..services.config import get_dataset_configs
from ..utils.helpers.common import should_refresh_cache
from ..utils.constants import CACHE_DIR

_API_BASE_URL = 'https://register.geonorge.no/api/det-offentlige-kartgrunnlaget-kommunalt.json?municipality='
_CACHE_DAYS = 7


def get_dataset_type(config: DatasetConfig) -> Literal['wfs', 'arcgis', 'ogc_api']:
    if config.wfs is not None:
        return 'wfs'
    elif config.arcgis is not None:
        return 'arcgis'
    elif config.ogc_api is not None:
        return 'ogc_api'

    return None


async def get_dataset_ids(data: dict, municipality_number: str) -> Dict[UUID, bool]:
    include_chosen_dok = data.get('includeFilterChosenDOK', True)

    if include_chosen_dok:
        kartgrunnlag = await _get_kartgrunnlag(municipality_number)
    else:
        kartgrunnlag = []

    dataset_ids = _get_datasets_by_theme(data.get('theme'))
    datasets: Dict[UUID, bool] = {}

    for id in dataset_ids:
        should_analyze = len(
            kartgrunnlag) == 0 or id is None or id in kartgrunnlag
        datasets[id] = should_analyze

    return datasets


def _get_datasets_by_theme(theme: str) -> List[UUID]:
    dataset_configs = get_dataset_configs()
    dataset_ids = []

    for config in dataset_configs:
        themes = list(map(lambda theme: theme.lower(), config.themes))

        if theme is None or theme.lower() in themes:
            dataset_ids.append(config.dataset_id)

    return dataset_ids


async def _get_kartgrunnlag(municipality_number: str) -> List[str]:
    if municipality_number is None:
        return []

    file_path = Path(path.join(
        CACHE_DIR, 'dok-datasets', f'{municipality_number}.json'))

    if not file_path.exists() or should_refresh_cache(file_path, _CACHE_DAYS):
        file_path.parent.mkdir(parents=True, exist_ok=True)
        dataset_ids = await _fetch_dataset_ids(municipality_number)
        json_object = json.dumps(dataset_ids, indent=2)

        with file_path.open('w', encoding='utf-8') as file:
            file.write(json_object)

        return dataset_ids
    else:
        with file_path.open(encoding='utf-8') as file:
            dataset_ids = json.load(file)

        return dataset_ids


async def _fetch_dataset_ids(municipality_number: str) -> List[str]:
    response = await _fetch_kartgrunnlag(municipality_number)

    if response is None:
        return []

    contained_items: List[Dict] = response.get('containeditems', [])
    datasets: List[str] = []

    for dataset in contained_items:
        if dataset.get('ConfirmedDok') == 'JA' and dataset.get('dokStatus') == 'Godkjent':
            metadata_url: str = dataset.get('MetadataUrl')
            splitted = metadata_url.split('/')
            datasets.append(splitted[-1])

    return datasets


async def _fetch_kartgrunnlag(municipality_number: str) -> dict:
    try:
        url = _API_BASE_URL + municipality_number

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    return None

                return await response.json()
    except:
        return None


__all__ = ['get_dataset_type',    'get_dataset_ids']



from typing import List, Dict, Tuple
import aiohttp
from lxml import etree as ET
from osgeo import ogr, osr
from ..http_clients.wfs import query_wfs

_WFS_URL = 'https://wfs.geonorge.no/skwms1/wfs.administrative_enheter'


async def get_municipality(geometry: ogr.Geometry, epsg: int) -> Tuple[str, str]:
    municipality = await _get_municipality_from_rest_api(geometry, epsg)

    if municipality is not None:
        return municipality

    return await _get_municipality_from_wfs(geometry, epsg)


async def _get_municipality_from_rest_api(geometry: ogr.Geometry, epsg: int) -> Tuple[str, str]:
    centroid: ogr.Geometry = geometry.Centroid()
    point: List[float] = centroid.GetPoint(0)

    return await _fetch_municipality(point[0], point[1], epsg)


async def _get_municipality_from_wfs(geometry: ogr.Geometry, epsg: int) -> Tuple[str, str]:
    centroid: ogr.Geometry = geometry.Centroid()
    spatial_ref: osr.SpatialReference = geometry.GetSpatialReference()
    centroid.AssignSpatialReference(spatial_ref)

    _, response = await query_wfs(_WFS_URL, 'Kommune', 'område', centroid, epsg, 5)

    if response is None:
        return None

    ns = {'wfs': 'http://www.opengis.net/wfs/2.0',
          'app': 'https://skjema.geonorge.no/SOSI/produktspesifikasjon/AdmEnheter/20240101'}

    root = ET.fromstring(bytes(response, encoding='utf-8'))
    municipality_number = root.findtext(
        './/wfs:member/*/app:kommunenummer', namespaces=ns)
    municipality_name = root.findtext(
        './/wfs:member/*/app:kommunenavn', namespaces=ns)

    return municipality_number, municipality_name


async def _fetch_municipality(x: float, y: float, epsg: int) -> Tuple[str, str]:
    try:
        url = f'https://api.kartverket.no/kommuneinfo/v1/punkt?nord={y}&ost={x}&koordsys={epsg}&filtrer=kommunenummer,kommunenavn'

        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=5) as response:
                if response.status != 200:
                    return None

                json: Dict = await response.json()

                municipality_number = json.get('kommunenummer', None)
                municipality_name = json.get('kommunenavn', None)

                return municipality_number, municipality_name
    except:
        return None


__all__ = ['get_municipality']


from abc import ABC, abstractmethod
from typing import List, Dict
from uuid import UUID
from osgeo import ogr
from .quality_measurement import QualityMeasurement
from .metadata import Metadata
from .exceptions import DokAnalysisException
from .result_status import ResultStatus
from .config.dataset_config import DatasetConfig
from .config.quality_indicator_type import QualityIndicatorType
from ..utils.helpers.common import keys_to_camel_case
from ..utils.helpers.geometry import create_buffered_geometry, create_run_on_input_geometry_json
from ..utils.helpers.map_image import create_payload_for_analysis
from ..utils.constants import MAP_IMAGE_API_URL
from ..services.config import get_quality_indicator_configs
from ..services.kartkatalog import get_kartkatalog_metadata
from ..services.quality.coverage_quality import get_coverage_quality
from ..services.quality.dataset_quality import get_dataset_quality
from ..services.quality.object_quality import get_object_quality
from ..services.map_image import create_map_image

_QMS_SORT_ORDER = [
    'fullstendighet_dekning',
    'stedfestingsnøyaktighet',
    'egnethet_reguleringsplan',
    'egnethet_kommuneplan',
    'egnethet_byggesak'
]


class Analysis(ABC):
    def __init__(self, dataset_id: UUID, config: DatasetConfig, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int):
        self.dataset_id = dataset_id
        self.config = config
        self.geometry = geometry
        self.run_on_input_geometry: ogr.Geometry = None
        self.epsg = epsg
        self.orig_epsg = orig_epsg
        self.geometries: List[ogr.Geometry] = []
        self.geolett: Dict = None
        self.title: str = None
        self.description: str = None
        self.guidance_text: str = None
        self.guidance_uri: List[Dict] = []
        self.possible_actions: List[str] = []
        self.quality_measurement: List[QualityMeasurement] = []
        self.quality_warning: List[str] = []
        self.buffer = buffer or 0
        self.input_geometry_area: ogr.Geometry = None
        self.run_on_input_geometry_json: Dict = None
        self.hit_area: float = None
        self.distance_to_object: int = 0
        self.raster_result_map: str = None
        self.raster_result_image: str = None
        self.raster_result_image_bytes: bytes = None
        self.cartography: str = None
        self.data: List[Dict] = []
        self.themes: List[str] = None
        self.run_on_dataset: Metadata = None
        self.run_algorithm: List[str] = []
        self.result_status: ResultStatus = ResultStatus.NO_HIT_GREEN
        self.coverage_statuses: List[str] = []
        self.has_coverage: bool = True

    async def run(self, context, include_guidance, include_quality_measurement) -> None:
        self.__set_input_geometry()

        await self.__run_coverage_analysis()

        if self.has_coverage:
            await self._run_queries()

            if self.result_status == ResultStatus.TIMEOUT or self.result_status == ResultStatus.ERROR:
                await self.set_default_data()
                return

            self.__set_geometry_areas()
        else:
            self.result_status = ResultStatus.NO_HIT_YELLOW

        if self.result_status in [ResultStatus.NO_HIT_GREEN, ResultStatus.NO_HIT_YELLOW]:
            await self._set_distance_to_object()

        self._add_run_algorithm('deliver result')

        self.run_on_input_geometry_json = create_run_on_input_geometry_json(
            self.run_on_input_geometry, self.epsg, self.orig_epsg)

        await self.set_default_data()

        if include_guidance and self.geolett is not None:
            self.__set_guidance_data()

        if include_quality_measurement:
            await self.__set_quality_measurements(context)

        if self.raster_result_map and MAP_IMAGE_API_URL:
            payload = create_payload_for_analysis(
                self.geometry, self.buffer, self.raster_result_map)
            _, result = await create_map_image(payload)
            self.raster_result_image_bytes = result

    def _add_run_algorithm(self, algorithm) -> None:
        self.run_algorithm.append(algorithm)

    async def set_default_data(self) -> None:
        self.title = self.geolett.get(
            'tittel') if self.geolett else self.config.title
        self.themes = self.config.themes
        self.run_on_dataset = await get_kartkatalog_metadata(self.config.metadata_id)

    async def __run_coverage_analysis(self) -> None:
        quality_indicators = get_quality_indicator_configs(self.dataset_id)

        if len(quality_indicators) == 0:
            return

        coverage_indicators = [
            indicator for indicator in quality_indicators if indicator.type == QualityIndicatorType.COVERAGE]

        if len(coverage_indicators) == 0:
            return
        elif len(coverage_indicators) > 1:
            raise DokAnalysisException(
                'A dataset can only have one coverage quality indicator')

        ci = coverage_indicators[0]

        self._add_run_algorithm(f'check coverage {ci.wfs.url}')
        measurements, warning, has_coverage = await get_coverage_quality(ci, self.run_on_input_geometry, self.epsg)
        self._add_run_algorithm(f'intersects layer {ci.wfs.layer} ({has_coverage})')

        self.quality_measurement.extend(measurements)
        self.has_coverage = has_coverage

        if warning is not None:
            self.quality_warning.append(warning)

    def __set_input_geometry(self) -> None:
        self._add_run_algorithm('set input_geometry')

        if self.buffer > 0:
            buffered_geom = create_buffered_geometry(
                self.geometry, self.buffer, self.epsg)
            self._add_run_algorithm(f'add buffer ({self.buffer})')
            self.run_on_input_geometry = buffered_geom
        else:
            self.run_on_input_geometry = self.geometry.Clone()

    def __set_geometry_areas(self) -> None:
        self.input_geometry_area = round(
            self.run_on_input_geometry.GetArea(), 2)

        if len(self.geometries) == 0:
            return

        hit_area: float = 0

        for geometry in self.geometries:
            if geometry is None:
                continue

            intersection: ogr.Geometry = self.run_on_input_geometry.Intersection(
                geometry)

            if intersection is None:
                continue

            geom_type = intersection.GetGeometryType()

            if geom_type == ogr.wkbPolygon or geom_type == ogr.wkbMultiPolygon:
                hit_area += intersection.GetArea()

        self._add_run_algorithm('calculate hit area')
        self.hit_area = round(hit_area, 2)

    def __set_guidance_data(self) -> None:
        if self.result_status != ResultStatus.NO_HIT_GREEN:
            self.description = self.geolett.get('forklarendeTekst')
            self.guidance_text = self.geolett.get('dialogtekst')

        for link in self.geolett.get('lenker', []):
            self.guidance_uri.append({
                'href': link['href'],
                'title': link['tittel']
            })

        possible_actions: str = self.geolett.get('muligeTiltak', '')

        for line in possible_actions.splitlines():
            self.possible_actions.append(line.lstrip('- '))

    async def __set_quality_measurements(self, context) -> None:
        quality_indicators = get_quality_indicator_configs(self.dataset_id)

        if len(quality_indicators) == 0:
            return

        dataset_qms, dataset_warnings = await get_dataset_quality(self.dataset_id, quality_indicators, context=context, themes=self.themes)
        object_qms, object_warnings = get_object_quality(
            quality_indicators, self.data)

        self.quality_measurement.extend(dataset_qms)
        self.quality_measurement.extend(object_qms)
        self.quality_warning.extend(dataset_warnings)
        self.quality_warning.extend(object_warnings)

    def __sort_quality_measurements(self) -> List[QualityMeasurement]:
        qms: List[QualityMeasurement] = []

        for id in _QMS_SORT_ORDER:
            result = [
                qm for qm in self.quality_measurement if qm.quality_dimension_id == id]

            if len(result) > 0:
                qms.extend(result)

        return qms

    def to_dict(self) -> Dict:
        sorted_qms = self.__sort_quality_measurements()

        return {
            'title': self.title,
            'runOnInputGeometry': self.run_on_input_geometry_json,
            'buffer': self.buffer,
            'runAlgorithm': self.run_algorithm,
            'inputGeometryArea': self.input_geometry_area,
            'hitArea': self.hit_area,
            'resultStatus': self.result_status,
            'distanceToObject': self.distance_to_object,
            'rasterResult': {
                'imageUri': self.raster_result_image,
                'mapUri': self.raster_result_map
            },
            'cartography': self.cartography,
            'data': list(map(lambda entry: keys_to_camel_case(entry), self.data)),
            'themes': self.themes,
            'runOnDataset': self.run_on_dataset.to_dict() if self.run_on_dataset is not None else None,
            'description': self.description,
            'guidanceText': self.guidance_text,
            'guidanceUri': self.guidance_uri,
            'possibleActions': self.possible_actions,
            'qualityMeasurement': list(map(lambda item: item.to_dict(), sorted_qms)),
            'qualityWarning': self.quality_warning
        }

    @abstractmethod
    async def _run_queries(self) -> None:
        pass

    @abstractmethod
    def _set_distance_to_object(self) -> None:
        pass



from sys import maxsize
from io import BytesIO
from typing import List, Dict
from osgeo import ogr
from lxml import etree as ET
from uuid import UUID
from .analysis import Analysis
from .result_status import ResultStatus
from .config.dataset_config import DatasetConfig
from .config.layer import Layer
from ..services.geolett import get_geolett_data
from ..services.raster_result import get_wms_url, get_cartography_url
from ..utils.helpers.common import parse_string, evaluate_condition, xpath_select_one
from ..utils.helpers.geometry import create_buffered_geometry, geometry_from_gml
from ..http_clients.wfs import query_wfs


class WfsAnalysis(Analysis):
    def __init__(self, dataset_id: UUID, config: DatasetConfig, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int):
        super().__init__(dataset_id, config, geometry, epsg, orig_epsg, buffer)

    async def _run_queries(self) -> None:
        first_layer = self.config.layers[0]
        geolett_data = await get_geolett_data(first_layer.geolett_id)
        self._add_run_algorithm(f'query {self.config.wfs}')

        for layer in self.config.layers:
            if layer.filter is not None:
                self._add_run_algorithm(f'add filter {layer.filter}')

            status_code, api_response = await query_wfs(
                self.config.wfs, layer.wfs, self.config.geom_field, self.run_on_input_geometry, self.epsg)

            if status_code == 408:
                self.result_status = ResultStatus.TIMEOUT
                self._add_run_algorithm(f'intersects layer {layer.wfs} (Timeout)')
                break
            elif status_code != 200:
                self.result_status = ResultStatus.ERROR
                self._add_run_algorithm(f'intersects layer {layer.wfs} (Error)')
                break

            if api_response:
                response = self.__parse_response(api_response, layer)

                if len(response['properties']) > 0:
                    self._add_run_algorithm(f'intersects layer {layer.wfs} (True)')
                    geolett_data = await get_geolett_data(layer.geolett_id)

                    self.geometries = response['geometries']
                    self.data = response['properties']
                    self.raster_result_map = get_wms_url(
                        self.config.wms, layer.wms)
                    self.cartography = await get_cartography_url(
                        self.config.wms, layer.wms)
                    self.result_status = layer.result_status
                    break

                self._add_run_algorithm(f'intersects layer {layer.wfs} (False)')            

        self.geolett = geolett_data

    async def _set_distance_to_object(self) -> None:
        buffered_geom = create_buffered_geometry(
            self.geometry, 20000, self.epsg)
        layer = self.config.layers[0]

        _, response = await query_wfs(self.config.wfs, layer.wfs, self.config.geom_field, buffered_geom, self.epsg)

        if response is None:
            self.distance_to_object = maxsize
            return

        source = BytesIO(response.encode('utf-8'))
        context = ET.iterparse(source, huge_tree=True)
        distances = []

        for _, elem in context:
            localname = ET.QName(elem).localname

            if localname != self.config.geom_field:
                continue

            geom_elem = xpath_select_one(elem, './*')

            if geom_elem is None:
                continue

            gml_str = ET.tostring(geom_elem, encoding='unicode')
            feature_geom = geometry_from_gml(gml_str)

            if feature_geom:
                distance = round(
                    self.run_on_input_geometry.Distance(feature_geom))
                distances.append(distance)

        distances.sort()
        self._add_run_algorithm('get distance to nearest object')

        if len(distances) == 0:
            self.distance_to_object = maxsize
        else:
            self.distance_to_object = distances[0]

    def __parse_response(self, wfs_response: str, layer: Dict) -> Dict[str, List]:
        data = {
            'properties': [],
            'geometries': []
        }

        source = BytesIO(wfs_response.encode('utf-8'))
        context = ET.iterparse(source, huge_tree=True)

        for _, elem in context:
            localname = ET.QName(elem).localname

            if localname != 'member':
                continue

            props = self.__map_properties(elem)

            if self.__filter_member(props, layer):
                data['properties'].append(props)
                data['geometries'].append(
                    self.__get_geometry_from_response(elem))

        return data

    def __filter_member(self, props: Dict, layer: Layer) -> bool:
        if not layer.filter:
            return True

        return evaluate_condition(layer.filter, props)

    def __map_properties(self, member: ET._Element) -> Dict:
        props = {}

        for mapping in self.config.properties:
            path = f'.//*[local-name() = "{mapping}"]/text()'
            value = xpath_select_one(member, path)

            if value:
                prop_name = mapping
                props[prop_name] = parse_string(value)

        return props

    def __get_geometry_from_response(self, member: ET._Element) -> ogr.Geometry:
        geom_field = self.config.geom_field
        path = f'.//*[local-name() = "{geom_field}"]/*'
        geom_elem = xpath_select_one(member, path)

        if geom_elem is None:
            return None

        gml_str = ET.tostring(geom_elem, encoding='unicode')

        return geometry_from_gml(gml_str)



import uuid
from typing import Optional, List, Dict
from pydantic import BaseModel, HttpUrl, root_validator
from .layer import Layer


class DatasetConfig(BaseModel):
    dataset_id: Optional[uuid.UUID] = None
    metadata_id: Optional[uuid.UUID] = None
    name: Optional[str] = None
    title: Optional[str] = None
    disabled: Optional[bool] = False
    wfs: Optional[HttpUrl] = None
    arcgis: Optional[HttpUrl] = None
    ogc_api: Optional[HttpUrl] = None
    wms: HttpUrl
    layers: List[Layer]
    geom_field: Optional[str] = None
    properties: Optional[List[str]]
    themes: List[str]

    @root_validator(pre=True)
    def check_service_type(cls, values: Dict) -> Dict:
        if not 'wfs' in values and not 'arcgis' in values and not 'ogc_api' in values:
            raise ValueError(
                'The dataset must have either the "wfs", "arcgis" or "ogc_api" property set')

        return values



from typing import List, Dict, Tuple
from osgeo import ogr
from . import get_threshold_values
from ..coverage import get_values_from_wfs
from ..codelist import get_codelist
from ...models.quality_measurement import QualityMeasurement
from ...models.config.quality_indicator import QualityIndicator


async def get_coverage_quality(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[List[QualityMeasurement], str, bool]:
    quality_data, has_coverage = await _get_coverage_quality_data(quality_indicator, geometry, epsg)

    if quality_data is None:
        return [], None, False

    measurements: List[QualityMeasurement] = []

    for value in quality_data.get('values'):
        measurements.append(QualityMeasurement(quality_data.get('id'), quality_data.get(
            'name'), value.get('value'), value.get('comment')))

    warning = quality_data.get('warning_text')

    return measurements, warning, has_coverage


async def _get_coverage_quality_data(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[Dict[str, any], bool]:
    values, hit_area_percent = await _get_values_from_web_service(quality_indicator, geometry, epsg)

    if len(values) == 0:
        return None, False

    codelist = await get_codelist('fullstendighet_dekning')
    meas_values: List[Dict] = []

    for value in values:
        meas_value = 'Nei' if value in [
            'ikkeKartlagt', 'ikkeRelevant'] else 'Ja'

        comment = _get_label_from_codelist(value, codelist)

        meas_values.append({
            'value': meas_value,
            'comment': comment
        })

    measurement = {
        'id': quality_indicator.quality_dimension_id,
        'name': quality_indicator.quality_dimension_name,
        'values': meas_values,
        'warning_text': _get_warning_text(quality_indicator, values, hit_area_percent)
    }

    return measurement, _has_coverage(values)


async def _get_values_from_web_service(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float]:
    if quality_indicator.wfs is not None:
        return await get_values_from_wfs(quality_indicator.wfs, geometry, epsg)

    # TODO: Add support for ArcGIS and OGC Features API

    return [], 0


def _get_warning_text(quality_indicator: QualityIndicator, values: List[str], hit_area_percent: float) -> str:
    threshold_values = get_threshold_values(quality_indicator)

    should_warn = any(value for value in values if any(
        t_value for t_value in threshold_values if t_value == value))

    warning_text = None

    if should_warn:
        warning_text: str = quality_indicator.quality_warning_text

        if 0 < hit_area_percent < 100:
            hit_area = str(hit_area_percent).replace('.', ',')
            warning_text = f'{hit_area} % av {warning_text.lower()}'

    return warning_text


def _has_coverage(values: List[str]) -> bool:
    if 'ikkeKartlagt' in values:
        has_other_values = any(
            value != 'ikkeKartlagt' for value in values)
        return has_other_values

    return True


def _get_label_from_codelist(value: str, codelist: List[Dict]) -> str:
    if codelist is None or len(codelist) == 0:
        return None

    result = next(
        (entry for entry in codelist if entry['value'] == value), None)

    return result.get('label') if result is not None else None


__all__ = ['get_coverage_quality']


from typing import Dict, Tuple
import asyncio
from osgeo import ogr, osr
from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError
from .services import analyses
from .utils.helpers.request import request_is_valid
from .utils.socket_io import get_client
from .utils import logger

logger.setup()

osr.UseExceptions()
ogr.UseExceptions()

PROCESS_METADATA = {
    'version': '0.1.0',
    'id': 'dokanalyse',
    'title': {
        'no': 'DOK-analyse'
    },
    'description': {
        'no': 'Tjeneste som utfører en en standardisert DOK-arealanalyse for enhetlig DOK-analyse på tvers av kommuner og systemleverandørerviser.',
    },
    'keywords': [
        'dokanalyse',
        'DOK'
    ],
    'links': [],
    'inputs': {
        'inputGeometry': {
            'title': 'Område',
            'description': 'Området en ønsker å analysere mot. Kan f.eks. være en eiendom eller et planområde.',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            },
            'minOccurs': 1,
            'maxOccurs': 1
        },
        'requestedBuffer': {
            'title': 'Ønsket buffer',
            'description': 'Antall meter som legges på inputGeometry som buffer i analysen. Kan utelates og avgjøres av analysen hva som er fornuftig buffer.',
            'schema': {
                'type': 'string'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'context': {
            'title': 'Kontekst',
            'description': 'Hint om hva analysen skal brukes til.',
            'schema': {
                'type': 'string'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'theme': {
            'title': 'Tema',
            'description': 'DOK-tema kan angis for å begrense analysen til aktuelle tema.',
            'schema': {
                'type': 'string'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'includeGuidance': {
            'title': 'Inkluder veiledning',
            'description': 'Velg om veiledningstekster skal inkluderes i resultat om det finnes i Geolett-registeret. Kan være avhengig av å styres med context for å få riktige tekster.',
            'schema': {
                'type': 'boolean'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'includeQualityMeasurement': {
            'title': 'Inkluder kvalitetsinformasjon',
            'description': 'Velg om kvalitetsinformasjon skal tas med i resultatet der det er mulig, slik som dekningskart, egnethet, nøyaktighet, osv.',
            'schema': {
                'type': 'boolean'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'includeFilterChosenDOK': {
            'title': 'Inkluder kun kommunens valgte DOK-data',
            'schema': {
                'type': 'boolean'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'includeFactSheet': {
            'title': 'Inkluder faktainformasjon',
            'schema': {
                'type': 'boolean'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        }
    },
    'outputs': {
        'resultList': {
            'title': 'Resultatliste',
            'description': 'Strukturert resultat på analysen',
            'schema': {
                'type': 'Result',
            },
            'minOccurs': 0
        },
        'report': {
            'title': 'Rapport',
            'description': 'Rapporten levert som PDF',
            'schema': {
                'type': 'binary',
            },
            'minOccurs': 0,
            'maxOccurs': 1
        },
        'inputGeometry': {
            'title': 'Område',
            'description': 'Valgt område for analyse',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            },
            'minOccurs': 0,
            'maxOccurs': 1
        }
    },
    'example': {
        'inputs': {
            'inputGeometry': {
                'type': 'Polygon',
                'coordinates': [
                    [
                        [
                            504132.67,
                            6585575.94
                        ],
                        [
                            504137.07,
                            6585483.64
                        ],
                        [
                            504286.5,
                            6585488.04
                        ],
                        [
                            504273.32,
                            6585575.94
                        ],
                        [
                            504132.67,
                            6585575.94
                        ]
                    ]
                ],
                'crs': {
                    'type': 'name',
                    'properties': {
                        'name': 'urn:ogc:def:crs:EPSG::25832'
                    }
                }
            },
            'requestedBuffer': 50,
            'context': 'Reguleringsplan',
            'theme': 'Natur',
            'includeGuidance': True,
            'includeQualityMeasurement': True,
            'includeFilterChosenDOK': True,
            'includeFactSheet': True
        }
    }
}


class DokanalyseProcessor(BaseProcessor):
    def __init__(self, processor_def):
        super().__init__(processor_def, PROCESS_METADATA)

    def execute(self, data: Dict, outputs=None) -> Tuple[str, Dict]:
        mimetype = 'application/json'

        if not request_is_valid(data):
            raise ProcessorExecuteError('Invalid payload')

        sio_client = get_client()

        try:
            outputs = asyncio.run(analyses.run(data, sio_client))
        finally:
            if sio_client:
                sio_client.disconnect()

        return mimetype, outputs or None

    def __repr__(self) -> str:
        return f'<DokanalyseProcessor> {self.name}'



import base64
from io import BytesIO
from typing import List
import asyncio
import aiohttp
from PIL import Image


async def create_legend(urls) -> str:
    tasks: List[asyncio.Task] = []

    async with asyncio.TaskGroup() as tg:
        for url in urls:
            tasks.append(tg.create_task(_fetch_image(url)))

    img_data: List[bytes] = []

    for task in tasks:
        img_data.append(task.result())

    return _merge_images(img_data)


def _merge_images(img_data: List[bytes]) -> str:
    imgs = [Image.open(BytesIO(i)) for i in img_data]
    max_width = max(i.width for i in imgs)
    total_height = sum(i.height for i in imgs)

    img_merge = Image.new('RGBA', (max_width, total_height), (255, 0, 0, 0))
    y = 0

    for img in imgs:
        img_merge.paste(img, (0, y))
        y += img.height

    buffered = BytesIO()
    img_merge.save(buffered, format='PNG')
    img_str = base64.b64encode(buffered.getvalue())

    return f'data:image/png;base64,{img_str.decode("ascii")}'


async def _fetch_image(url) -> bytes:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    return None

                return await response.read()
    except:
        return None


__all__ = ['create_legend']



import time
import logging
import traceback
from typing import List, Dict
from uuid import UUID, uuid4
import asyncio
from socketio import SimpleClient
from osgeo import ogr
from .dataset import get_dataset_ids, get_dataset_type
from .fact_sheet import create_fact_sheet
from .municipality import get_municipality
from ..services.config import get_dataset_config
from ..services.blob_storage import create_container, upload_image
from ..utils.helpers.geometry import create_input_geometry, get_epsg
from ..models.config import DatasetConfig
from ..models import Analysis, ArcGisAnalysis, OgcApiAnalysis, WfsAnalysis, EmptyAnalysis, AnalysisResponse, ResultStatus
from ..utils.constants import DEFAULT_EPSG
from ..utils.correlation_id_middleware import get_correlation_id

_LOGGER = logging.getLogger(__name__)


async def run(data: Dict, sio_client: SimpleClient) -> AnalysisResponse:
    geo_json = data.get('inputGeometry')
    geometry = create_input_geometry(geo_json)
    orig_epsg = get_epsg(geo_json)
    buffer = data.get('requestedBuffer', 0)
    context = data.get('context')
    include_guidance = data.get('includeGuidance', False)
    include_quality_measurement = data.get('includeQualityMeasurement', False)
    include_facts = data.get('includeFacts', True)
    municipality_number, municipality_name = await get_municipality(geometry, DEFAULT_EPSG)

    datasets = await get_dataset_ids(data, municipality_number)
    correlation_id = get_correlation_id()

    if datasets and correlation_id and sio_client:
        to_analyze = {key: value for (
            key, value) in datasets.items() if value == True}
        sio_client.emit('datasets_counted_api', {'count': len(
            to_analyze), 'recipient': correlation_id})

    tasks: List[asyncio.Task] = []

    async with asyncio.TaskGroup() as tg:
        for dataset_id, should_analyze in datasets.items():
            task = tg.create_task(_run_analysis(
                dataset_id, should_analyze, geometry, DEFAULT_EPSG, orig_epsg, buffer,
                context, include_guidance, include_quality_measurement, sio_client))
            tasks.append(task)

    fact_sheet = None

    if include_facts:
        if correlation_id and sio_client:
            sio_client.emit('create_fact_sheet_api', {
                            'recipient': correlation_id})

        fact_sheet = await create_fact_sheet(geometry, orig_epsg, buffer)

    response = AnalysisResponse.create(
        geo_json, geometry, DEFAULT_EPSG, orig_epsg, buffer, fact_sheet, municipality_number, municipality_name)

    for task in tasks:
        response.result_list.append(task.result())

    await _upload_binaries(response)

    return response.to_dict()


async def _run_analysis(dataset_id: UUID, should_analyze: bool, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int,
                        context: str, include_guidance: bool, include_quality_measurement: bool, sio_client: SimpleClient) -> Analysis:
    config = get_dataset_config(dataset_id)

    if config is None:
        return None

    if not should_analyze:
        analysis = EmptyAnalysis(
            config.dataset_id, config, ResultStatus.NOT_RELEVANT)
        await analysis.run()
        return analysis

    start = time.time()
    correlation_id = get_correlation_id()

    analysis = _create_analysis(
        dataset_id, config, geometry, epsg, orig_epsg, buffer)

    try:
        await analysis.run(context, include_guidance, include_quality_measurement)
    except Exception:
        err = traceback.format_exc()
        _LOGGER.error(err)
        await analysis.set_default_data()
        analysis.result_status = ResultStatus.ERROR

    end = time.time()

    # autopep8: off
    _LOGGER.info(f'Dataset analyzed: {dataset_id} - {config.name}: {round(end - start, 2)} sec.')
    # autopep8: on

    if correlation_id and sio_client:
        sio_client.emit('dataset_analyzed_api', {
            'dataset': str(dataset_id), 'recipient': correlation_id})

    return analysis


def _create_analysis(dataset_id: UUID, config: DatasetConfig, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int) -> Analysis:
    dataset_type = get_dataset_type(config)

    match dataset_type:
        case 'arcgis':
            return ArcGisAnalysis(dataset_id, config, geometry, epsg, orig_epsg, buffer)
        case 'ogc_api':
            return OgcApiAnalysis(dataset_id, config, geometry, epsg, orig_epsg, buffer)
        case 'wfs':
            return WfsAnalysis(dataset_id, config, geometry, epsg, orig_epsg, buffer)
        case _:
            return None


async def _upload_binaries(response: AnalysisResponse):
    map_images: Dict[str, bytes] = {}

    if response.fact_sheet and response.fact_sheet.raster_result_image_bytes:
        map_images['fact_sheet'] = response.fact_sheet.raster_result_image_bytes

    for analysis in response.result_list:
        if analysis.raster_result_image_bytes:
            map_images[str(analysis.dataset_id)
                       ] = analysis.raster_result_image_bytes

    if not map_images:
        return

    container_name = str(uuid4())
    await create_container(container_name)

    tasks: List[asyncio.Task[str]] = []

    async with asyncio.TaskGroup() as tg:
        for key, value in map_images.items():
            blob_name = f'{str(uuid4())}.png'
            task = tg.create_task(upload_image(
                value, container_name, blob_name), name=key)
            tasks.append(task)

    for task in tasks:
        task_name = task.get_name()

        if task_name == 'fact_sheet':
            response.fact_sheet.raster_result_image = task.result()
            continue

        analysis = _find_analysis(response.result_list, task_name)

        if analysis:
            analysis.raster_result_image = task.result()


def _find_analysis(analyses: List[Analysis], dataset_id: str) -> Analysis:
    return next((analysis for analysis in analyses if str(analysis.dataset_id) == dataset_id), None)


__all__ = ['run']



from os import path
import logging
from typing import Tuple
from pydantic import HttpUrl
import aiohttp
import asyncio
from osgeo import ogr

_LOGGER = logging.getLogger(__name__)


async def query_wfs(url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Tuple[int, str]:
    gml_str = geometry.ExportToGML(['FORMAT=GML3'])
    request_xml = _create_wfs_request_xml(layer, geom_field, gml_str, epsg)

    return await _query_wfs(url, request_xml, timeout)


def _create_wfs_request_xml(layer: str, geom_field: str, gml_str: str, epsg: int) -> str:
    dir_path = path.dirname(path.realpath(__file__))
    file_path = path.join(dir_path, 'wfs_request.xml.txt')

    with open(file_path, 'r') as file:
        file_text = file.read()

    return file_text.format(layer=layer,  geom_field=geom_field, geometry=gml_str, epsg=epsg).encode('utf-8')


async def _query_wfs(url: HttpUrl, xml_body: str, timeout: int) -> Tuple[int, str]:
    url = f'{url}?service=WFS&version=2.0.0'
    headers = {'Content-Type': 'application/xml'}

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=xml_body, headers=headers, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                return 200, await response.text()
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)        
        return 500, None


__all__ = ['query_wfs']



import json
from typing import Dict, List
from osgeo import ogr, osr
from math import pi
from re import search
from shapely import wkt
from shapely.wkt import dumps
from ..constants import DEFAULT_EPSG, WGS84_EPSG

_EARTH_RADIUS = 6371008.8


def geometry_from_gml(gml_str: str) -> ogr.Geometry:
    try:
        return ogr.CreateGeometryFromGML(gml_str)
    except:
        return None


def geometry_from_json(json_str: str) -> ogr.Geometry:
    try:
        return ogr.CreateGeometryFromJson(json_str)
    except:
        return None


def geometry_to_wkt(geometry: ogr.Geometry, epsg: int) -> str:
    wkt_str = geometry.ExportToWkt()
    geometry = wkt.loads(wkt_str)
    coord_precision = 6 if epsg == WGS84_EPSG else 2

    return dumps(geometry, trim=True, rounding_precision=coord_precision)


def geometry_to_arcgis_geom(geometry: ogr.Geometry, epsg: int) -> str:
    if geometry.GetGeometryType() == ogr.wkbMultiPolygon:
        out_geom = ogr.ForceToPolygon(geometry)
    else:
        out_geom = geometry

    coord_precision = 6 if epsg == WGS84_EPSG else 2
    geojson = out_geom.ExportToJson(
        [f'COORDINATE_PRECISION={coord_precision}'])
    obj = json.loads(geojson)

    arcgis_geom = {
        'rings': obj['coordinates'],
        'spatialReference': {
            'wkid': epsg
        }
    }

    return json.dumps(arcgis_geom)


def create_input_geometry(geo_json: Dict) -> ogr.Geometry:
    epsg = get_epsg(geo_json)
    geometry = ogr.CreateGeometryFromJson(str(geo_json))

    if epsg != DEFAULT_EPSG:
        return transform_geometry(geometry, epsg, DEFAULT_EPSG)

    return geometry


def create_buffered_geometry(geometry: ogr.Geometry, distance: int, epsg: int) -> ogr.Geometry:
    computed_buffer = length_to_degrees(
        distance) if epsg is None or epsg == WGS84_EPSG else distance

    return geometry.Buffer(computed_buffer, 10)


def create_feature_collection(features: List[Dict], epsg: int = 4326) -> Dict:
    feature_collection = {
        'type': 'FeatureCollection',
        'features': features
    }

    add_geojson_crs(feature_collection, epsg)

    return feature_collection


def create_feature(geometry: ogr.Geometry, properties: Dict = {}) -> Dict:
    json_str = geometry.ExportToJson()

    return {
        'type': 'Feature',
        'geometry': json.loads(json_str),
        'properties': properties
    }


def transform_geometry(geometry: ogr.Geometry, src_epsg: int, dest_epsg: int) -> ogr.Geometry:
    source = osr.SpatialReference()
    source.ImportFromEPSG(src_epsg)
    source.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    target = osr.SpatialReference()
    target.ImportFromEPSG(dest_epsg)

    transform = osr.CoordinateTransformation(source, target)
    clone: ogr.Geometry = geometry.Clone()
    clone.Transform(transform)

    return clone


def length_to_degrees(distance: float) -> float:
    radians = distance / _EARTH_RADIUS
    degrees = radians % (2 * pi)

    return degrees * 180 / pi


def create_run_on_input_geometry_json(geometry: ogr.Geometry, epsg: int, orig_epsg: int) -> Dict:
    geom = geometry

    if epsg != orig_epsg:
        geom = transform_geometry(geometry, epsg, orig_epsg)

    coord_precision = 6 if orig_epsg == WGS84_EPSG else 2
    geo_json = json.loads(geom.ExportToJson(
        [f'COORDINATE_PRECISION={coord_precision}']))
    add_geojson_crs(geo_json, epsg)

    return geo_json


def get_epsg(geo_json: Dict) -> int:
    crs = geo_json.get('crs', {}).get('properties', {}).get('name')

    if crs is None:
        return WGS84_EPSG

    regex = r'^(http:\/\/www\.opengis\.net\/def\/crs\/EPSG\/0\/|^urn:ogc:def:crs:EPSG::|^EPSG:)(?P<epsg>\d+)$'
    matches = search(regex, crs)

    if matches:
        return int(matches.group('epsg'))

    return WGS84_EPSG


def get_epsg_from_geometry(geometry: ogr.Geometry):
    sr: osr.SpatialReference = geometry.GetSpatialReference()
    epsg: int = sr.GetAuthorityCode(None)

    return epsg or 4326


def add_geojson_crs(geojson: Dict, epsg: int) -> None:
    if epsg is None or epsg == WGS84_EPSG:
        return

    geojson['crs'] = {
        'type': 'name',
        'properties': {
            'name': 'urn:ogc:def:crs:EPSG::' + str(epsg)
        }
    }


__all__ = [
    'geometry_from_gml',
    'geometry_from_json',
    'geometry_to_wkt',
    'geometry_to_arcgis_geom',
    'create_input_geometry',
    'create_buffered_geometry',
    'create_feature_collection',
    'create_feature'
    'length_to_degrees',
    'create_run_on_input_geometry_json',
    'get_epsg',
    'get_epsg_from_geometry',
    'add_geojson_crs'
]
